"""GCP class file."""

import datetime
import json
import logging
import os

from google.cloud import exceptions, storage
from google.oauth2 import service_account
from google.oauth2.service_account import Credentials

from cloud_bucket_adapter.custom_blob import CustomBlob


class GCP(object):
    """GCP.

    Args:
        object ([type]): [description]

    Returns:
        [object]: [custom blob/resource object]
    """

    REQUIRED_FIELDS = ['CREDENTIAL_FILE', 'PROJECT_NAME', 'BUCKET_NAME']

    def authenticate(self, options):
        """[authenticate function to authenticate the service (aws s3/gcp bucket)].

        Args:
            options ([dict]): [options dict contains all the configuration settings]

        Returns:
            [object]: [a client object when authentication is successful else exception is raised]
        """
        try:
            credentials = None
            if 'CREDENTIAL_JSON' in options and options['CREDENTIAL_JSON']:
                credentials = service_account.Credentials.from_service_account_info(
                    json.loads(options['CREDENTIAL_JSON']))
            else:
                credentials = service_account.Credentials.from_service_account_file(
                    options['CREDENTIAL_FILE'])
            if credentials:
                client = storage.Client(
                    credentials=credentials, project=options['PROJECT_NAME'])
            else:
                raise exceptions.Forbidden('Authentication Failed')
            if client:
                return client
            else:
                raise exceptions.GoogleCloudError('Authentication Failed')
        except Exception as E:
            logging.exception("Exception {err}".format())
            raise Exception('Authentication Failed')

    def upload(self, filename, options, client, bucket_filename=None, ExtraArgs=None):
        """upload function to uplaod the file to gcp bucket.

        Args:
            filename ([string]): [file to be uploaded to bucket]
            options ([dict]): [options dict contains all the configuration settings]

        Returns:
            [tuple]: [bool success and uploaded file bucket url ]
        """
        try:
            if bucket_filename is None:
                bucket_filename = filename
            bucket = client.get_bucket(options['BUCKET_NAME'])
            blob = bucket.blob(bucket_filename)
            blob.upload_from_filename(filename)
            return True, blob.public_url
        except Exception as E:
            logging.error("Exception {err}".format(err=str(E)))

    def download(self, filename, options, client):
        """[download function to download the file in your working directory].

        Args:
            filename ([string]): [file to be download]
            options ([dict]): [dict object containing all the configuration settings]
            client ([object]): [client object received after successful authentication]

        Returns:
            [file]: [file is being downloaded in the working directory]
        """
        bucket = client.get_bucket(options['BUCKET_NAME'])
        blob = bucket.get_blob(filename)
        # FIXME: storage_name not defined?
        return blob.download_to_filename(storage_name)

    def generate_signed_url(self, filename, options, client):
        """Generate a v2 signed URL for downloading a blob.

        Note that this method requires a service account key file. You can not use
        this if you are using Application Default Credentials from Google Compute
        Engine or from the Google Cloud SDK.


        Args:
            file_name ([string]): [filename to be downloaded from bucket]
            options ([dict]): [options dict contains all the configuration settings]

        Returns:
            [string]: [a signed URL from where you can download the file and see it content]
        """
        bucket = client.bucket(options['BUCKET_NAME'])
        blob = bucket.blob(blob_name=filename)
        url = blob.generate_signed_url(
            expiration=datetime.timedelta(hours=1),
            method="GET",
        )
        return url

    def get_blob(self, filename, options, client):
        """[get custom blob object].

        Args:
            filename ([string]): [filename]
            options ([dict]): [options dict contains all the configuration settings]
            client ([object]): [client object received after successful authentication]

        Returns:
            [object]: [a custom blob object]
        """
        bucket = client.get_bucket(options['BUCKET_NAME'])
        blob = bucket.get_blob(filename)
        blob = CustomBlob(blob=blob, options=options)
        return blob

    def generate_signed_url_with_custom_expiry(self, filename, client, expiration, options):
        """[generate_signed_url_with_custom_expiry by passing the file expiry time as well].

        Args:
            filename ([string]): [file to be ]
            client ([object]): [client object received after successful authentication]
            expiration ([string]): [custom expiry to be set for the signed url]
            options ([dict]): [options dict contains all the configuration settings]

        Returns:
            [string]: [a signed url]
        """
        bucket = client.bucket(options['BUCKET_NAME'])
        blob = bucket.blob(blob_name=filename)
        url = blob.generate_signed_url(
            expiration=expiration,
            method="GET",
        )
        return url

    def download_to_file_pointer(self, filename, tempfile_name, client, options):
        """download to file pointer.

        Args:
            filename ([string]): [description]
            client ([object]): [client object of the service used (aws s3/gcp bucket)]
            options ([dict]): [options dict that contains all the configuration settings]

        Returns:
            [type]: [temporary file pointer object]
        """
        bucket = client.get_bucket(options['BUCKET_NAME'])
        blob = bucket.get_blob(filename)
        response = client.download_blob_to_file(blob, tempfile_name)
        return response

    def get_head_object(self, Key, options, client):
        """get head_object of GCP cloud storage object.

        Args:
            Key ([string]): [file name]
            options ([dict]): [description]
            client ([dict]): [description]
        """
        # FIXME
        raise NotImplementedError('Not Implemented on GCP')



"""adapter class."""

import import_string


class Adapter(object):
    """Adapter.

    Args:
        object ([object]): [main adapter class which calls the adaptee functions]
    """

    def __init__(self, settings_file):
        """__init__ function.

        Args:
            settings_file ([type]): [main configuration settings of the respective service being used (s3/gcp)]
        """
        self.settings = settings_file
        adaptee_classname = self.settings['NAME']
        mod = import_string(adaptee_classname)
        self.adaptee_obj = mod()
        self.authenticate = self.adaptee_obj.authenticate(
            options=self.settings)

    def upload(self, *args, **kwargs):
        """upload.

        Returns:
            [string]: [returns the url afer uploading the file to bucket]
        """
        return self.adaptee_obj.upload(
            *args, **kwargs, options=self.settings, client=self.authenticate)

    def download(self, *args, **kwargs):
        """download.

        Returns:
            [file]: [downlaod the file in the working directory]
        """
        return self.adaptee_obj.download(*args, **kwargs, options=self.settings, client=self.authenticate)

    def generate_signed_url(self, *args, **kwargs):
        """generate the signed url.

        Returns:
            [string]: [a signed url which expires in an hour]
        """
        return self.adaptee_obj.generate_signed_url(*args, **kwargs, options=self.settings, client=self.authenticate)

    def get_blob(self, *args, **kwargs):
        """give us a custom blob object.

        Returns:
            [object]: [a custom blob/resource object]
        """
        return self.adaptee_obj.get_blob(*args, **kwargs, options=self.settings, client=self.authenticate)

    def generate_signed_url_with_custom_expiry(self, *args, **kwargs):
        """generate the signed url with custom expiry.

        Returns:
            [string]: [returns the signed url with custom expiry]
        """
        return self.adaptee_obj.generate_signed_url_with_custom_expiry(*args, **kwargs, options=self.settings, client=self.authenticate)

    def download_to_file_pointer(self, *args, **kwargs):
        """download_to_file_pointer.

        Returns:
            [type]: [returns the file pointer]
        """
        return self.adaptee_obj.download_to_file_pointer(*args, **kwargs, options=self.settings, client=self.authenticate)
    
    def get_head_object(self, *args, **kwargs):
        """get_head_object.

        Returns:
            [type]: [returns the head object]
        """
        return self.adaptee_obj.get_head_object(*args, **kwargs, options=self.settings, client=self.authenticate)



"""custom blob."""


class CustomBlob(object):
    """CustomBlob to make a generic blob for both s3 & gcp bucket.

    Args:
        object ([type]): [description]
    """

    def __init__(self, blob, options, filename=None):
        """__init__ function to call the custom blob/resource function by determining the service type.

        Args:
            blob ([class object]): [blob object type s3/gcp]
            options ([dict]): [options dict contains all the configuration settings]
            filename ([string], optional): [filename used in s3 resource as it dont have any property like gcp blob.name and to make a custom & generic resource we need to add it.]. Defaults to None.
        """
        self.name = None
        self.time_created = None
        self.bucket = None
        self.content_type = None
        self.content_encoding = None
        self.content_language = None
        self.size = None
        print("TYPE:", type(blob), "is equal", )
        if str(type(blob)) == "<class 'google.cloud.storage.blob.Blob'>":
            self._generic_gcp_blob(blob, options)
        # Quick Hack because amazon uses generic factory functions to create
        # resource objects.
        elif str(type(blob)) == "<class 'boto3.resources.factory.s3.Object'>":
            self._generic_aws_resource(blob, options, filename=filename)
        # else:
        #     # custom function call
        #     func_pointer = options.get('BLOB_ETL_FUNCTION')
        #     func_pointer = import_string(func_pointer)
        #     func_pointer(self)

    def _generic_gcp_blob(self, blob, options, filename=None):
        """_generic_gcp_blob private function for custom gcp blob object.

        Args:
            blob ([object]): [blob object]
            options ([dict]): [options dict contains all the configuration settings]
            filename ([string], optional): [not used here as we have blob.name option in gcp blob object]. Defaults to None.
        """
        self.name = blob.name
        self.time_created = blob.time_created
        self.bucket = blob.bucket
        self.content_encoding = blob.content_encoding
        self.content_language = blob.content_language
        self.content_type = blob.content_type
        self.size = blob.size

    def _generic_aws_resource(self, blob, options, filename=None):
        """generic aws resource private function for custom s3 resource object.

        Args:
            blob ([object]): [resource object]
            options ([dict]): [options dict contains all the configuration settings]
            filename ([string], optional): [filename to be used in resource.name]. Defaults to None.
        """
        self.name = filename
        self.time_created = blob.last_modified
        self.bucket = options['BUCKET_NAME']
        self.content_encoding = blob.content_encoding
        self.content_language = blob.content_language
        self.content_type = blob.content_type
        self.size = blob.content_length



"""AWS adapter."""

import logging
from datetime import datetime, timezone

import boto3
from botocore.config import Config
from botocore.exceptions import ClientError
from dateutil.relativedelta import relativedelta

from cloud_bucket_adapter.custom_blob import CustomBlob


class AWS(object):
    """Main AWS adapter class.

    Args:
        object ([type]): [adaptee function to be called]
    """

    # Default signing version to be used.
    DEFAULT_VERSION = 's3v4'

    def _get_signature_version(self, options: dict) -> str:
        """return a valid signature version for use in signing.

        Args:
            options (dict): options dict contains all the configuration settings

        Returns:
            str: Defaults to s3v4 if not found in options.
        """
        version = options.get('SIGNATURE_VERSION', self.DEFAULT_VERSION)

    def authenticate(self, options):
        """authenticate function to authenticate the service (aws s3/gcp bucket).

        Args:
            options ([dict]): [options dict contains all the configuration settings]

        Returns:
            [object]: [a client object when authentication is successful else exception is raised]
        """
        try:
            client = boto3.client('s3',
                                  config=Config(
                                      signature_version=self._get_signature_version(options)),
                                  **options['CREDENTIALS'])
            return client
        except ClientError as e:
            logging.error(e)
            raise Exception('Authentication Failed')

    def upload(self, filename, options, client, bucket_filename=None, ExtraArgs=None):
        """Upload a file to an S3 bucket.

        :param file_name: File to upload
        :param bucket: Bucket to upload to
        :param object_name: S3 object name. If not specified then file_name is used
        :return: True if file was uploaded, else False
        """
        if bucket_filename is None:
            bucket_filename = filename
        try:
            client.upload_file(
                filename, options['BUCKET_NAME'], bucket_filename, ExtraArgs)
            # NOTE: S3 client does not return anything, so to make sure our interface
            # signature remains same, we will add a dummy url with file name.
            url_link = f'https://localhost/{bucket_filename}'
            return True, url_link
        except ClientError as e:
            logging.error(e)
            return False

    def download(self, filename, options, client, bucket_filename=None):
        """download file.

        Args:
            filename ([type]): [description]
            options ([type]): [description]
            object_name ([type], optional): [description]. Defaults to None.

        Returns:
            [type]: [description]
        """
        try:
            if bucket_filename is None:
                bucket_filename = filename
            response = client.download_file(
                options['BUCKET_NAME'], bucket_filename, filename)
            # FIXME: return proper type.
            return ('File Downloaded in your working directory')
        except ClientError as e:
            logging.error(e)
            return None

    def generate_signed_url(self, filename, options, client, expiration=3600):
        """signed url with expiry of 1 hour.

        Args:
            filename ([string]): [description]
            options ([dict]): [options dict contains all the configuration settings]
            client ([object]): [a client object received after successful authentication]
            expiration (int, optional): [expiry of 1 hour by default]. Defaults to 3600.

        Returns:
            [type]: [description]
        """
        try:
            response = client.generate_presigned_url('get_object',
                                                     Params={'Bucket': options['BUCKET_NAME'],
                                                             'Key': filename},
                                                     ExpiresIn=expiration)
            return response
        except ClientError as e:
            logging.error(e)
            return None

    def get_blob(self, filename, options, client=None):
        """a resource object.

        Args:
            filename ([string]): [filename]
            options ([dict]): [options dict contains all the configuration settings]

        Returns:
            [object]: [a custom blob/resource object]
        """
        client = boto3.resource('s3',
                                aws_access_key_id=options['ACCESS_KEY'],
                                aws_secret_access_key=options['SECRET_KEY'],
                                config=Config(
                                    signature_version=self._get_signature_version(options))
                                )
        resource = client.Object(options['BUCKET_NAME'], filename)
        resource = CustomBlob(
            blob=resource, options=options, filename=filename)
        return resource

    def generate_signed_url_with_custom_expiry(self, filename, client, expiration, options):
        """signed url with custom expiry.

        Args:
            filename ([string]): [filenam to generate signed url]
            client ([object]): [a client object received after successful authentication]
            expiration ([datettime]): [custom expiry for the signed url]
            options ([dict]): [options dict contains all the configuration settings]

        Returns:
            [string]: [a signed url of the file you want to access (download)]
        """
        try:
            if not expiration:
                # defaulting to an hour
                expiration = datetime.utcnow().replace(tzinfo=timezone.utc) + \
                    relativedelta(seconds=3600)
            delta = expiration - datetime.utcnow().replace(tzinfo=timezone.utc)
            response = client.generate_presigned_url('get_object',
                                                     Params={'Bucket': options['BUCKET_NAME'],
                                                             'Key': filename},
                                                     ExpiresIn=delta.seconds)
            return response
        except ClientError as e:
            logging.error(e)
            return None

    def download_to_file_pointer(self, filename, tempfile_name, client, options):
        """download to file pointer.

        Args:
            filename ([string]): [description]
            client ([object]): [client object of the service used (aws s3/gcp bucket)]
            options ([dict]): [options dict that contains all the configuration settings]

        Returns:
            [type]: [description]
        """
        resource = boto3.resource('s3',
                                  aws_access_key_id=options['ACCESS_KEY'],
                                  aws_secret_access_key=options['SECRET_KEY'],
                                  config=Config(
                                      signature_version=self._get_signature_version(options))
                                  )
        bucket = resource.Bucket(options['BUCKET_NAME'])
        response = bucket.download_fileobj(filename, tempfile_name)
        return response

    def get_head_object(self, filename, options, client):
        """get head_object of s3 object.

        Args:
            Key ([string]): [file name]
            options ([dict]): [description]
            client ([dict]): [description]

        Returns:
            [dict]: [meta data of S3 object]
        """
        try:
            response = client.head_object(
                Bucket=options['BUCKET_NAME'], Key=filename)
            return True, response
        except ClientError as e:
            logging.error(e)
            return False, None
